@@ -40,6 +40,9 @@ void slabs_stats(ADD_STAT add_stats, void *c);
 /* Hints as to freespace in slab class */
 unsigned int slabs_available_chunks(unsigned int id, bool *mem_flag, uint64_t *total_bytes, unsigned int *chunks_perslab);
 
+void slabs_mlock(void);
+void slabs_munlock(void);
+
 int start_slab_maintenance_thread(void);
 void stop_slab_maintenance_thread(void);
 
